/*
This is a part of the IconDesk source code.

Copyright (C) 2003 Erik Christiansson, aka Sci
www.alphafish.com
erik@alphafish.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "Group.h"

Group::Group(std::string sNameX, Settings* pSettings, HWND parentWnd)
{
	LogPrintf(LOG_DEBUG, "Group \"%s\"", sNameX.c_str());
	// Store input
	sName = sNameX;
	_pSettings = pSettings;
	_parentWnd = parentWnd;

	// We have not loaded our icons yet
	_iconsLoaded = FALSE;

	// A general buffer used to read the settings
	std::string sBuffer;

	// Get group metrics
	_visible = _pSettings->GetBool("Visible", true);
	_iconX = _pSettings->GetInt("X", 0);
	_iconY = _pSettings->GetInt("Y", 0);
	if(_iconX < 0) _iconX += GetSystemMetrics(SM_CXSCREEN);
	if(_iconY < 0) _iconY += GetSystemMetrics(SM_CYSCREEN);
	_iconSize = _pSettings->GetInt("Size", 32);
	_iconSpacingX = _pSettings->GetInt("SpacingX", 0);
	_iconSpacingY = _pSettings->GetInt("SpacingY", 0);

	_iconHideExtension = _pSettings->GetBool("HideExtension", true);
	_iconFoldersFirst = _pSettings->GetBool("FoldersFirst", true);
	_iconNonFilesystemFirst = _pSettings->GetBool("NonFilesystemFirst", true);
	_iconShowHidden = _pSettings->GetBool("ShowHidden", false);
	_showTooltips = _pSettings->GetBool("Tooltips", true);

	// Get text settings
	_hideText = _pSettings->GetBool("HideText", false);
	if(!_hideText) {
		// Get the background brush
		sBuffer = _pSettings->GetString("TextBackground", "");
		if(sBuffer.size()) {
			_textBackground = CreateSolidBrush(_pSettings->GetColor("TextBackground", RGB(255,255,255)));
		}
		else {
			_textBackground = NULL;
		}

		// Text box metrics
		_textSpacing = _pSettings->GetInt("TextSpacing", 5);
		_textHeight = _pSettings->GetInt("TextHeight", 28);
		_textWidth = _pSettings->GetInt("TextWidth", 64);
	
		// Relative position of the text box
		sBuffer = _pSettings->GetString("TextPosition", "");
		if(!stricmp(sBuffer, "right")) {
			_textPosition = DD_RIGHT;
		}
		else if(!stricmp(sBuffer, "left")) {
			_textPosition = DD_LEFT;
		}
		else if(!stricmp(sBuffer, "top")) {
			_textPosition = DD_UP;
		}
		else {
			_textPosition = DD_DOWN;
		}
	
		// Text alignment
		sBuffer = _pSettings->GetString("TextAlign", "");
		if(!stricmp(sBuffer, "right")) {
			_textAlign = DT_RIGHT;
		}
		else if(!stricmp(sBuffer, "left")) {
			_textAlign = DT_LEFT;
		}
		else {
			_textAlign = DT_CENTER;
		}

		// Get the font
		_textColor = _pSettings->GetColor("FontColor", RGB(255, 255, 255));
		sBuffer = _pSettings->GetString("Font", "Arial");
		int fontSize = _pSettings->GetInt("FontSize", 14);
		BOOL fontItalic = _pSettings->GetBool("FontItalic", FALSE);
		BOOL fontBold = _pSettings->GetBool("FontBold", FALSE);
		_textFont = CreateFont(fontSize, 0, 0, 0, fontBold ? FW_BOLD : FW_NORMAL, fontItalic, FALSE, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, sBuffer.c_str());
		if(!_textFont) {
			LogPrintf(LOG_ERROR, "Could not load font \"%s\"", sBuffer.c_str());
		}

		// Text shadow settings
		_textShadowColor = _pSettings->GetColor("TextShadowColor", RGB(80, 80, 80));
		_textShadowDepth = _pSettings->GetInt("TextShadowDepth", 0);

		// Text outline settings
		_textOutlineColor = _pSettings->GetColor("TextOutlineColor", RGB(40, 40, 40));
		_textOutline = _pSettings->GetBool("TextOutline", false);
		if(_textOutline) {
			if(_textShadowDepth > 0) {
				_textShadowDepth++;
			}
			else if(_textShadowDepth < 0) {
				_textShadowDepth--;
			}
		}
	}
	else {
		_textWidth = 0;
		_textHeight = 0;
		_textSpacing = 0;
	}

	// Icon direction
	sBuffer = _pSettings->GetString("Direction", "");
	if(!stricmp(sBuffer, "down")) {
		_iconDirection = DD_DOWN;
	}
	else if(!stricmp(sBuffer, "left")) {
		_iconDirection = DD_LEFT;
	}
	else if(!stricmp(sBuffer, "up")) {
		_iconDirection = DD_UP;
	}
	else {
		_iconDirection = DD_RIGHT;
	}

	// Icon wraping
	sBuffer = _pSettings->GetString("WrapDirection", "");
	if(_iconDirection == DD_DOWN || _iconDirection == DD_UP) {
		if(!stricmp(sBuffer, "left")) {
			_iconWrapDirection = DD_LEFT;
		}
		else {
			_iconWrapDirection = DD_RIGHT;
		}
	}
	else {
		if(!stricmp(sBuffer, "up")) {
			_iconWrapDirection = DD_UP;
		}
		else {
			_iconWrapDirection = DD_DOWN;
		}
	}
	_iconWrapCount = max(_pSettings->GetInt("WrapCount", 10), 1);

	// Set the default mouse actions
	UINT nMask;
	nMask = ME_LBUTTONDBL;
	_mouseEvents.insert(MouseMapValue(nMask, ".open"));
	nMask |= ME_MENU;
	_mouseEvents.insert(MouseMapValue(nMask, ".properties"));
	nMask = ME_RBUTTONUP;
	_mouseEvents.insert(MouseMapValue(nMask, ".context"));

	// Read mouse actions
	for(Settings::iterator it = _pSettings->begin("Mouse"); it != _pSettings->end(); ++it) {
		unsigned int nStart = 0;
		unsigned int nLength = 0;
		std::string sLine = (*it);
		std::string sBlank = " \t";
		std::string sEvent;
		std::string sModifiers;
		std::string sCommand;

		// Get event
		nLength = sLine.find_first_of(sBlank, 0);
		if(nLength == std::string::npos || nLength == 0) {
			break;
		}
		sEvent = sLine.substr(0, nLength);
		nStart = nLength;

		// Get modifiers
		nStart = sLine.find_first_not_of(sBlank, nStart);
		if(nStart != std::string::npos) {
			break;
		}
		nLength = sLine.find_first_of(sBlank, nStart);
		if(nLength == std::string::npos) {
			break;
		}
		nLength -= nStart;
		if(nLength == 0) {
			break;
		}
		sModifiers = sLine.substr(nStart, nLength);
		nStart += nLength;

		// Get command
		nStart = sLine.find_first_not_of(sBlank, nStart);
		if(nStart == std::string::npos) {
			break;
		}
		nLength = sLine.size() - nStart;
		if(nLength == 0) {
			break;
		}
		sCommand = sLine.substr(nStart, nLength);
		nStart += nLength;

		UINT nMask = 0;

		// Parse mouse events
		if(!stricmp(sEvent, "LeftUp")) {
			nMask |= ME_LBUTTONUP;
		}
		else if(!stricmp(sEvent, "LeftDown")) {
			nMask |= ME_LBUTTONDOWN;
		}
		else if(!stricmp(sEvent, "LeftDbl")) {
			nMask |= ME_LBUTTONDBL;
		}
		else if(!stricmp(sEvent, "MiddleUp")) {
			nMask |= ME_MBUTTONUP;
		}
		else if(!stricmp(sEvent, "MiddleDown")) {
			nMask |= ME_MBUTTONDOWN;
		}
		else if(!stricmp(sEvent, "MiddleDbl")) {
			nMask |= ME_MBUTTONDBL;
		}
		else if(!stricmp(sEvent, "RightUp")) {
			nMask |= ME_RBUTTONUP;
		}
		else if(!stricmp(sEvent, "RightDown")) {
			nMask |= ME_RBUTTONDOWN;
		}
		else if(!stricmp(sEvent, "RightDbl")) {
			nMask |= ME_RBUTTONDBL;
		}
		else {
			// Notify the user that an invalid mouse event was found
			ErrorMessage("'%s' is not a valid mouse event.", sEvent.c_str());
		}

		// Parse modifier keys
		std::string sModifier;
		nStart = 0;
		nLength = 0;
		while(nStart < sModifiers.size()) {
			nLength = sModifiers.find_first_of("+", nStart);
			if(nLength == std::string::npos) {
				nLength = sModifiers.size();
			}
			nLength -= nStart;
			sModifier = sModifiers.substr(nStart, nLength);
			nStart += nLength + 1;

			if(!stricmp(sModifier, "CTRL")) {
				nMask |= ME_CONTROL;
			}
			else if(!stricmp(sModifier, "ALT")) {
				nMask |= ME_MENU;
			}
			else if(!stricmp(sModifier, "SHIFT")) {
				nMask |= ME_SHIFT;
			}
			else if(!stricmp(sModifier, "APPS")) {
				nMask |= ME_APPS;
			}
			else if(stricmp(sModifier, ".none")) {
				// Notify the user that an invalid modifier was found
				ErrorMessage("'%s' is not a valid modifier key.", sModifier.c_str());
			}
		}

		MouseMapIter it;
		it = _mouseEvents.find(nMask);
		if(it != _mouseEvents.end()) {
			_mouseEvents.erase(it);
		}
		_mouseEvents.insert(MouseMapValue(nMask, sCommand));
	}

	// Icon effects
	_effectFlags = 0;
	_hueIntensity = (UCHAR)min(max(_pSettings->GetInt("HueIntensity", 0), 0), 255);
	_hueColor = _pSettings->GetColor("HueColor", RGB(128, 128, 128));
	if(_hueIntensity) {
		_effectFlags |= FX_HUE;
	}
	_saturation = (UCHAR)min(max(_pSettings->GetInt("Saturation", 255), 0), 255);
	if(_saturation != 255) {
		_effectFlags |= FX_SATURATION;
	}

	// Hide patterns
	for(Settings::iterator it = _pSettings->begin("Hide"); it != _pSettings->end(); ++it) {
		_patternHide.push_back((*it));
	}

	// First patterns
	for(Settings::iterator it = _pSettings->begin("First"); it != _pSettings->end(); ++it) {
		// Add pattern to the list
		_patternFirst.push_back((*it));
	}

	// Last patterns
	for(Settings::iterator it = _pSettings->begin("Last"); it != _pSettings->end(); ++it) {
		// Add pattern to the list
		_patternLast.push_back((*it));
	}

	// Load the folders
	for(Settings::iterator it = _pSettings->begin("Folder"); it != _pSettings->end(); ++it) {
		// Remove prefix and whitespace
		std::string sData = (*it);

		// Remove quotes around the path
		if(sData[0] == '"' && sData[sData.size() - 1] == '"') {
			sData = sData.substr(1, sData.size() - 2);
		}

		_folders.push_back(sData);

		LogPrintf(LOG_DEBUG, "Folder \"%s\" \"%s\"", sName.c_str(), sData.c_str());

		// Folder watcher
		LPITEMIDLIST pidl = NULL;
		IShellFolder *psfDesktop = NULL;
		HRESULT hr;
		WCHAR wszFolder[MAX_PATH];

		if(!stricmp(sData, ".desktop")) {
			LPMALLOC pMalloc;
			if(SUCCEEDED(SHGetMalloc(&pMalloc))) {
				void *temp = pMalloc->Alloc(2);
				pMalloc->Release();
				ZeroMemory(temp, 2);
				pidl = (LPITEMIDLIST)temp;
			}
		}
		else {
			// Get the IShellFolder for the folder
			MultiByteToWideChar(CP_ACP, 0, sData.c_str(), -1, wszFolder, sizeof(wszFolder) / sizeof(WCHAR));
			// Used to bind the interface
			hr = SHGetDesktopFolder(&psfDesktop);
			// Get the PIDL
			hr = psfDesktop->ParseDisplayName(NULL, NULL, wszFolder, NULL, &pidl, NULL);
			// Release the desktop
			psfDesktop->Release();
		}
		_notifiers.push_back(new ShellNotify(pidl, this));
	}

	// If no folder was specified, use the desktop
	if(_folders.size() == 0) {
		// The default folder is the desktop
		_folders.push_back(".desktop");

		// Create a folder watcher
		LPITEMIDLIST pidl;
		LPMALLOC pMalloc;
		if(SUCCEEDED(SHGetMalloc(&pMalloc))) {
			void *temp = pMalloc->Alloc(2);
			pMalloc->Release();
			ZeroMemory(temp, 2);
			pidl = (LPITEMIDLIST)temp;

			LogPrintf(LOG_DEBUG, "Default folder \"%s\" \".desktop\"", sName.c_str());

			// Folder watcher
			_notifiers.push_back(new ShellNotify(pidl, this));
		}
	}
}

Group::~Group()
{
	_deleteIcons();

	if(_textFont) {
		DeleteObject(_textFont);
	}
	if(_textBackground) {
		DeleteObject(_textBackground);
	}

	for(ShellNotifyListIter it = _notifiers.begin(); it != _notifiers.end(); it++) {
		delete (*it);
	}
}

void Group::show()
{
	_loadTooltips();
	_visible = TRUE;
	InvalidateRect(_parentWnd, NULL, TRUE);
}

void Group::hide()
{
	_deleteTooltips();
	_visible = FALSE;
	InvalidateRect(_parentWnd, NULL, TRUE);
}

void Group::toggle()
{
	if(_visible) {
		hide();
	}
	else {
		show();
	}
}

void Group::update()
{
	_deleteIcons();
	_loadIcons();
	InvalidateRect(_parentWnd, NULL, TRUE);
}

BOOL Group::mouseEvent(UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	if(!_visible) {
		return FALSE;
	}

	Icon* clicked = _getIconAtPt(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));

	if(clicked) {
		if(uMsg == WM_RBUTTONDOWN || uMsg == WM_LBUTTONDOWN || uMsg == WM_MBUTTONDOWN) {
			if(_selectedIcon) {
				InvalidateRect(_parentWnd, &_selectedIcon->textRect, TRUE);
			}
			_selectedIcon = clicked;
			if(_selectedIcon) {
				InvalidateRect(_parentWnd, &_selectedIcon->textRect, TRUE);
			}
		}

		UINT nMask = 0;

		if(GetAsyncKeyState(VK_SHIFT) & 0x8000) {
			nMask |= ME_SHIFT;
		}

		if(GetAsyncKeyState(VK_CONTROL) & 0x8000) {
			nMask |= ME_CONTROL;
		}

		if(GetAsyncKeyState(VK_MENU) & 0x8000) {
			nMask |= ME_MENU;
		}

		if(GetAsyncKeyState(VK_APPS) & 0x8000) {
			nMask |= ME_APPS;
		}

		switch(uMsg) {
			case WM_LBUTTONUP:
				nMask |= ME_LBUTTONUP;
				break;

			case WM_LBUTTONDOWN:
				nMask |= ME_LBUTTONDOWN;
				break;

			case WM_LBUTTONDBLCLK:
				nMask |= ME_LBUTTONDBL;
				break;

			case WM_MBUTTONUP:
				nMask |= ME_MBUTTONUP;
				break;

			case WM_MBUTTONDOWN:
				nMask |= ME_MBUTTONDOWN;
				break;

			case WM_MBUTTONDBLCLK:
				nMask |= ME_MBUTTONDBL;
				break;

			case WM_RBUTTONUP:
				nMask |= ME_RBUTTONUP;
				break;

			case WM_RBUTTONDOWN:
				nMask |= ME_RBUTTONDOWN;
				break;

			case WM_RBUTTONDBLCLK:
				nMask |= ME_RBUTTONDBL;
				break;

		}

		MouseMapIter it;
		it = _mouseEvents.find(nMask);
		if(it != _mouseEvents.end()) {
			clicked->Execute((*it).second);
        }
		return TRUE;
	}
	else {
		if(_selectedIcon) {
			InvalidateRect(_parentWnd, &_selectedIcon->textRect, TRUE);
		}
		_selectedIcon = NULL;
		return FALSE;
	}
}

void Group::_deleteIcons()
{
	for(IconListIter it = _iconList.begin(); it != _iconList.end(); it++) {
		if((*it)->toolId) {
			DeleteTooltip((*it)->toolId);
		}

		delete (*it);
	}
	_iconList.clear();
	_iconsLoaded = FALSE;
}

void Group::_loadTooltips()
{
	if(_showTooltips) {
		for(IconListIter it = _iconList.begin(); it != _iconList.end(); it++) {
			(*it)->toolId = AddTooltip((*it)->sTitle, (*it)->iconRect, (*it)->toolId);
		}
	}
}

void Group::_deleteTooltips()
{
	if(_showTooltips) {
		for(IconListIter it = _iconList.begin(); it != _iconList.end(); it++) {
			if((*it)->toolId) {
				DeleteTooltip((*it)->toolId);
			}
		}
	}
}

void Group::_loadFolder(std::string sFolder)
{
    LPMALLOC pMalloc;
    LPITEMIDLIST pidlFolder = NULL;
    LPITEMIDLIST pidlItems = NULL;
    IShellFolder *psfDesktop = NULL;
    IShellFolder *psfFolder = NULL;
    LPENUMIDLIST ppenum = NULL;
    ULONG celtFetched;
    HRESULT hr;
    STRRET strDispName;
    ULONG uAttr;
	char szBuffer[MAX_PATH];
	WCHAR wszFolder[MAX_PATH];
	BOOL bFilesystem = TRUE;
	char szMyDocuments[1024];

	// For memory managment
    hr = SHGetMalloc(&pMalloc);

	if(!stricmp(sFolder, ".desktop")) {
	    hr = SHGetDesktopFolder(&psfFolder);
		// Create a pidl with only the terminating NULLs
		pidlFolder = (LPITEMIDLIST)pMalloc->Alloc(2);
		ZeroMemory(pidlFolder, 2);
		bFilesystem = FALSE;
	}
	else {
		// Get the IShellFolder for the folder
		MultiByteToWideChar(CP_ACP, 0, sFolder.c_str(), -1, wszFolder, sizeof(wszFolder) / sizeof(WCHAR));
		// Used to bind the interface
		hr = SHGetDesktopFolder(&psfDesktop);
		// Get the PIDL
		hr = psfDesktop->ParseDisplayName(NULL, NULL, wszFolder, NULL, &pidlFolder, NULL);
		// Get the interface
		hr = psfDesktop->BindToObject(pidlFolder, NULL, IID_IShellFolder, (LPVOID *)&psfFolder);
		// Is it a filesystem folder?
		uAttr = SFGAO_FILESYSTEM;
        psfDesktop->GetAttributesOf(1, (LPCITEMIDLIST*)&pidlFolder, &uAttr);
		if((uAttr & SFGAO_FILESYSTEM) == NULL) {
			bFilesystem = FALSE;
		}
		// Release the desktop
		psfDesktop->Release();
	}

	// Ugly hack for My Documents
	if(!bFilesystem) {
		DWORD dwDataSize = 1024;
		SHGetValue(HKEY_CURRENT_USER, "Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders", "Personal", NULL, szMyDocuments, &dwDataSize);
	}

	// Get the enumeration object
	DWORD enumFlags = SHCONTF_FOLDERS | SHCONTF_NONFOLDERS;
	if(_iconShowHidden) {
		enumFlags |= SHCONTF_INCLUDEHIDDEN;
	}
    hr = psfFolder->EnumObjects(NULL, enumFlags, &ppenum);
    while(ppenum->Next(1, &pidlItems, &celtFetched) == S_OK && (celtFetched) == 1) {
		// Create a icon object
		FileIcon* icon = new FileIcon();

		icon->hParent = _parentWnd;

		// Get the title
		uAttr = SFGAO_FILESYSTEM;
        psfFolder->GetAttributesOf(1, (LPCITEMIDLIST*)&pidlItems, &uAttr);
        psfFolder->GetDisplayNameOf(pidlItems, SHGDN_INFOLDER, &strDispName);
		_StrRetToBuf(&strDispName, pidlItems, szBuffer, MAX_PATH);
		icon->sTitle = szBuffer;

		// Get the filename
        psfFolder->GetDisplayNameOf(pidlItems, SHGDN_NORMAL | SHGDN_FORPARSING, &strDispName);
        _StrRetToBuf(&strDispName, pidlItems, szBuffer, MAX_PATH);
		icon->sFilename = szBuffer;

		// Get the full PIDL (relative to the desktop)
		icon->SetPIDL(PIDLAppend(pidlFolder, pidlItems, pMalloc));

		// Is it a folder?
		uAttr = SFGAO_FOLDER;
        psfFolder->GetAttributesOf(1, (LPCITEMIDLIST*)&pidlItems, &uAttr);
		if(uAttr & SFGAO_FOLDER) {
			icon->lFlags = icon->lFlags | Icon::ICON_FOLDER;
		}

		// Is it a filesystem object?
		uAttr = SFGAO_FILESYSTEM;
        psfFolder->GetAttributesOf(1, (LPCITEMIDLIST*)&pidlItems, &uAttr);
		if(uAttr & SFGAO_FILESYSTEM) {
			icon->lFlags = icon->lFlags | Icon::ICON_FILESYSTEM;
		}

		// Ugly hack to fix My Documents
		if(!bFilesystem) {
			if(!strcmp(icon->sFilename.c_str(), szMyDocuments)) {
				icon->sFilename = "::{450D8FBA-AD25-11D0-98A8-0800361B1103}";
				icon->lFlags = icon->lFlags ^ Icon::ICON_FILESYSTEM;
			}
		}

		// Hide extension
		if(_iconHideExtension && (icon->lFlags & Icon::ICON_FOLDER) == 0) {
			strncpy(szBuffer, icon->sTitle.c_str(), MAX_PATH);
			PathRemoveExtension(szBuffer);
			icon->sTitle = szBuffer;
		}

		// Release the PIDL
        pMalloc->Free(pidlItems);

		// Check for first patterns
		for(StringListIter first = _patternFirst.begin(); first != _patternFirst.end(); first++) {
			if(match((*first).c_str(), icon->sTitle.c_str()) || match((*first).c_str(), icon->sFilename.c_str())) {
				icon->lFlags |= Icon::ICON_FIRST;
				break;
			}
		}

		// Check for last patterns
		for(StringListIter last = _patternLast.begin(); last != _patternLast.end(); last++) {
			if(match((*last).c_str(), icon->sTitle.c_str()) || match((*last).c_str(), icon->sFilename.c_str())) {
				icon->lFlags |= Icon::ICON_LAST;
				break;
			}
		}

		// If the icon title matches any of the hide patterns, don't show it
		bool bHide = false;
		for(StringListIter hide = _patternHide.begin(); hide != _patternHide.end(); hide++) {
			bHide = match((*hide).c_str(), icon->sTitle.c_str()) || match((*hide).c_str(), icon->sFilename.c_str());
			if(bHide) {
				break;
			}
		}

		if(!bHide) {
			// Add it to the list
			_iconList.push_back((Icon*)icon);
		}
		else {
			delete icon;
		}
    }

	// Release interfaces/PIDLs
    ppenum->Release();
    pMalloc->Free(pidlFolder);
    pMalloc->Release();
    psfFolder->Release();
}

void Group::_loadIcons()
{
	_iconsLoaded = TRUE;

	for(StringListIter slit = _folders.begin(); slit != _folders.end(); slit++) {
		_loadFolder((*slit).c_str());
	}

	// Sort the icon list
	_iconList.sort(CompareIconFlags(NULL, false));
	if(_iconFoldersFirst) {
		_iconList.sort(CompareIconFlags(Icon::ICON_FOLDER, false));
	}
	if(_iconNonFilesystemFirst) {
		_iconList.sort(CompareIconFlags(Icon::ICON_FILESYSTEM, true));
	}
	_iconList.sort(CompareIconFlags(Icon::ICON_FIRST, false));
	_iconList.sort(CompareIconFlags(Icon::ICON_LAST, true));

	// Calculate RECT's and load icons
	int i = 0, x = 0, y = 0;
	int boxWidth = 0, boxHeight = 0;
	int offsetIconX = 0, offsetIconY = 0;
	int offsetTextX = 0, offsetTextY = 0;

	if(_hideText) {
		boxWidth = _iconSize;
		boxHeight = _iconSize;
		offsetIconX = 0;
		offsetIconY = 0;
		offsetTextX = 0;
		offsetTextY = 0;
	}
	else {
		switch(_textPosition) {
		case DD_DOWN:
			boxWidth = max(_iconSize, _textWidth);
			boxHeight = _iconSize + _textSpacing + _textHeight;
			offsetIconX = max(_iconSize, _textWidth) / 2 - _iconSize / 2;
			offsetIconY = 0;
			offsetTextX = max(_iconSize, _textWidth) / 2 - _textWidth / 2;
			offsetTextY = _iconSize + _textSpacing;
			break;

		case DD_UP:
			boxWidth = max(_iconSize, _textWidth);
			boxHeight = _iconSize + _textSpacing + _textHeight;
			offsetIconX = max(_iconSize, _textWidth) / 2 - _iconSize / 2;
			offsetIconY = _textHeight + _textSpacing;
			offsetTextX = max(_iconSize, _textWidth) / 2 - _textWidth / 2;
			offsetTextY = 0;
			break;

		case DD_RIGHT:
			boxWidth = _iconSize + _textSpacing + _textWidth;
			boxHeight = max(_iconSize, _textHeight);
			offsetIconX = 0;
			offsetIconY = max(_iconSize, _textHeight) / 2 - _iconSize / 2;
			offsetTextX = _iconSize + _textSpacing;
			offsetTextY = max(_iconSize, _textHeight) / 2 - _textHeight / 2;
			break;

		case DD_LEFT:
			boxWidth = _iconSize + _textSpacing + _textWidth;
			boxHeight = max(_iconSize, _textHeight);
			offsetIconX = _textWidth + _textSpacing;
			offsetIconY = max(_iconSize, _textHeight) / 2 - _iconSize / 2;
			offsetTextX = 0;
			offsetTextY = max(_iconSize, _textHeight) / 2 - _textHeight / 2;
			break;

		}
	}

	IconListIter icon;
	for(icon = _iconList.begin(); icon != _iconList.end(); icon++) {
		switch(_iconDirection) {
			case DD_RIGHT:
			case DD_LEFT:
				x = (i % _iconWrapCount);
				y = (i / _iconWrapCount);
				if(_iconWrapDirection == DD_UP) {
					y = -y;
				}
				if(_iconDirection == DD_LEFT) {
					x = -x;
				}
				break;

			case DD_DOWN:
			case DD_UP:
				x = (i / _iconWrapCount);
				y = (i % _iconWrapCount);
				if(_iconWrapDirection == DD_LEFT) {
					x = -x;
				}
				if(_iconDirection == DD_UP) {
					y = -y;
				}
				break;

		}

		if(!_hideText) {
			(*icon)->x = _iconX + ((boxWidth + _iconSpacingX) * x);
			(*icon)->y = _iconY + ((boxHeight + _iconSpacingY) * y);
			SetRect(&(*icon)->textRect, (*icon)->x + offsetTextX, (*icon)->y + offsetTextY, (*icon)->x + offsetTextX + _textWidth, (*icon)->y + offsetTextY + _textHeight);
			SetRect(&(*icon)->iconRect, (*icon)->x + offsetIconX, (*icon)->y + offsetIconY, (*icon)->x + offsetIconX + _iconSize, (*icon)->y + offsetIconY + _iconSize);
		}
		else {
			(*icon)->x = _iconX + ((_iconSize + _iconSpacingX) * x);
			(*icon)->y = _iconY + ((_iconSize + _iconSpacingY) * y);

			SetRect(&(*icon)->iconRect, (*icon)->x, (*icon)->y, (*icon)->x + _iconSize, (*icon)->y + _iconSize);
		}

		// Load icon
		(*icon)->LoadIcon(_iconSize);
		if(!(*icon)->GetIcon()) {
			LogPrintf(LOG_ERROR, "Could not load icon for \"%s\"", (*icon)->sTitle.c_str());
		}

		// Create buffer copy with fx changes entered
		if(_effectFlags & FX_ANY) {
			HICON hIcon = (*icon)->GetIcon();
			if(hIcon != NULL) {
				PAINTSTRUCT ps;
				HDC	hdcScreen = BeginPaint(_parentWnd, &ps);

				ICONINFO ii;
				GetIconInfo(hIcon, &ii);

				HDC hdcBuffer = CreateCompatibleDC(hdcScreen);
				SelectObject(hdcBuffer, ii.hbmColor);

				HDC hdc = CreateCompatibleDC(hdcBuffer);
				HBITMAP hbmNew = CreateCompatibleBitmap(hdcBuffer, 2 * _iconSize, _iconSize);
				SelectObject(hdc, hbmNew);

				// Temporarily draw up the icon to read it's pixels
				DrawIconEx(hdc, 0, 0, hIcon, _iconSize, _iconSize, 0, NULL, DI_NORMAL);
				// the mask of the icon
				DrawIconEx(hdc, _iconSize, 0, hIcon, _iconSize, _iconSize, 0, NULL, DI_MASK);

				int y = 0;
				int x = 0;
				COLORREF cl = 0;
				BYTE r = 0;
				BYTE g = 0;
				BYTE b = 0;
				while(y < _iconSize) {
					x = 0;
					while(x < _iconSize) {
						// Loop through all transparent pixels...
						while(x < _iconSize) {
							// If the mask-pixel is white, then break
							if(!GetPixel(hdc, x + _iconSize, y)) {
								cl = GetPixel(hdc, x, y);
								r = GetRValue(cl);
								g = GetGValue(cl);
								b = GetBValue(cl);

								// Saturation effect
								if(_effectFlags & FX_SATURATION) {
									BYTE gray = (BYTE)(r * 0.3086 + g * 0.6094 + b * 0.0820);

									r = (BYTE)((r * _saturation + gray * (255 - _saturation) + 255) >> 8);
									g = (BYTE)((g * _saturation + gray * (255 - _saturation) + 255) >> 8);
									b = (BYTE)((b * _saturation + gray * (255 - _saturation) + 255) >> 8);
								}

								// Hue effect
								if(_effectFlags & FX_HUE) {
									// The B & R values of the hue is swapped here, can somebody tell me why it only works that way?
									// test normal
									r = (BYTE)((r * (255 - _hueIntensity) + GetRValue(_hueColor) * _hueIntensity + 255) >> 8);
									g = (BYTE)((g * (255 - _hueIntensity) + GetGValue(_hueColor) * _hueIntensity + 255) >> 8);
									b = (BYTE)((b * (255 - _hueIntensity) + GetBValue(_hueColor) * _hueIntensity + 255) >> 8);
								}
								SetPixel(hdcBuffer, x, y, RGB(r, g, b));
							}
							x++;
						}
					}
					y++;
				}
	
				(*icon)->SetIcon(CreateIconIndirect(&ii));

				DeleteObject(hbmNew);
				DeleteDC(hdc);
				DeleteDC(hdcBuffer);
				DeleteDC(hdcScreen);
				EndPaint(_parentWnd, &ps);
			}
		}

		i++;
	}

	_loadTooltips();
}

void Group::paintIcons(HDC hDC)
{
	if(_visible) {
		if(_iconsLoaded == FALSE) {
			_loadIcons();
		}

		COLORREF oldColor = SetTextColor(hDC, _textColor);
		HFONT oldFont = (HFONT)SelectObject(hDC, _textFont);
		int oldMode = SetBkMode(hDC, TRANSPARENT);
		HICON hIcon = NULL;

		for(IconListIter it = _iconList.begin(); it != _iconList.end(); it++) {
			hIcon = (*it)->GetIcon();
			if(hIcon != NULL) {
				DrawIconEx(hDC, (*it)->iconRect.left, (*it)->iconRect.top, hIcon, _iconSize, _iconSize, 0, NULL, DI_NORMAL);
			}

			if(_hideText == false) {
				if(_textBackground) {
					FillRect(hDC, &(*it)->textRect, _textBackground);
				}

				if(_textShadowDepth != 0) {
					RECT r;
					CopyRect(&r, &(*it)->textRect);
					OffsetRect(&r, _textShadowDepth, _textShadowDepth);
					SetTextColor(hDC, _textShadowColor);
					DrawText(hDC, (*it)->sTitle.c_str(), (*it)->sTitle.size(), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);

					if(_textOutline == true) {
						OffsetRect(&r, -1, 0);
						DrawText(hDC, (*it)->sTitle.c_str(), (*it)->sTitle.size(), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
						OffsetRect(&r, -1, 0);
						DrawText(hDC, (*it)->sTitle.c_str(), (*it)->sTitle.size(), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
						OffsetRect(&r, 0, -1);
						DrawText(hDC, (*it)->sTitle.c_str(), (*it)->sTitle.size(), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
						OffsetRect(&r, 0, -1);
						DrawText(hDC, (*it)->sTitle.c_str(), (*it)->sTitle.size(), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
						OffsetRect(&r, 1, 0);
						DrawText(hDC, (*it)->sTitle.c_str(), (*it)->sTitle.size(), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
						OffsetRect(&r, 1, 0);
						DrawText(hDC, (*it)->sTitle.c_str(), (*it)->sTitle.size(), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
						OffsetRect(&r, 0, 1);
						DrawText(hDC, (*it)->sTitle.c_str(), (*it)->sTitle.size(), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
						SetTextColor(hDC, _textColor);
					}
					SetTextColor(hDC, _textColor);
				}

				if(_textOutline == true) {
					RECT r;
					CopyRect(&r, &(*it)->textRect);
					OffsetRect(&r, 1, 1);
					SetTextColor(hDC, _textOutlineColor);
					DrawText(hDC, (*it)->sTitle.c_str(), (*it)->sTitle.size(), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
					OffsetRect(&r, -1, 0);
					DrawText(hDC, (*it)->sTitle.c_str(), (*it)->sTitle.size(), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
					OffsetRect(&r, -1, 0);
					DrawText(hDC, (*it)->sTitle.c_str(), (*it)->sTitle.size(), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
					OffsetRect(&r, 0, -1);
					DrawText(hDC, (*it)->sTitle.c_str(), (*it)->sTitle.size(), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
					OffsetRect(&r, 0, -1);
					DrawText(hDC, (*it)->sTitle.c_str(), (*it)->sTitle.size(), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
					OffsetRect(&r, 1, 0);
					DrawText(hDC, (*it)->sTitle.c_str(), (*it)->sTitle.size(), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
					OffsetRect(&r, 1, 0);
					DrawText(hDC, (*it)->sTitle.c_str(), (*it)->sTitle.size(), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
					OffsetRect(&r, 0, 1);
					DrawText(hDC, (*it)->sTitle.c_str(), (*it)->sTitle.size(), &r, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);
					SetTextColor(hDC, _textColor);
				}

				DrawText(hDC, (*it)->sTitle.c_str(), (*it)->sTitle.size(), &(*it)->textRect, _textAlign | DT_WORDBREAK | DT_NOPREFIX | DT_END_ELLIPSIS);

				/*if(_selectedIcon == (*it)) {
					DrawFocusRect(hDC, &(*it)->textRect);
				}*/
			}
		}

		SetTextColor(hDC, oldColor);
		SelectObject(hDC, oldFont);
		SetBkMode(hDC, oldMode);
	}
}

Icon* Group::_getIconAtPt(int x, int y)
{
	for(IconListIter it = _iconList.begin(); it != _iconList.end(); it++) {
		if((*it)->PointOver(x, y)) {
			return (*it);
		}
	}
	return NULL;
}

void Group::_StrRetToBuf(LPSTRRET pstr, LPCITEMIDLIST pidl, LPTSTR pszBuf, UINT cchBuf)
{
	switch(pstr->uType) {
		case STRRET_WSTR:
			{
				WideCharToMultiByte(CP_ACP, 0, pstr->pOleStr, -1, pszBuf, cchBuf, NULL, NULL);
				LPMALLOC pMalloc;
				if(SUCCEEDED(SHGetMalloc(&pMalloc))) {
					pMalloc->Free(pstr->pOleStr);
					pMalloc->Release();
				}
			}
			break;

		case STRRET_CSTR:
			lstrcpyn(pszBuf, pstr->cStr, cchBuf);
			break;

		case STRRET_OFFSET:
			lstrcpyn(pszBuf, (LPTSTR)pidl + pstr->uOffset, cchBuf);
			break;

	}
}
